using System;
using System.Collections.Generic;
using System.IO;
using System.Windows.Forms;
using System.Xml;
using JetBrains.ReSharper.Daemon;
using JetBrains.ReSharper.ProjectModel;
using JetBrains.ReSharper.Psi;
using JetBrains.ReSharper.Psi.CSharp;
using JetBrains.ReSharper.Psi.CSharp.Tree;
using JetBrains.ReSharper.Psi.Resolve;
using JetBrains.ReSharper.Psi.Tree;
using JetBrains.ReSharper.TextControl;
using ReSharper.Framework;

namespace ReSharper.CatchExceptions {
  /// <summary>
  /// Implements a Generate Assertions powertoy plugin.
  /// </summary>
  [ContextAction(Name="CatchExceptions")]
  public class ContextAction : ContextActionBase, IComparer<Pair<string, Type>> {
    /// <summary>
    /// Tests this instance.
    /// </summary>
    public void Test() {
      File.Delete("C:\\test.cv");
    }

    #region Constructor

    /// <summary>
    /// Initializes a new instance of the <see cref="ContextAction"/> class.
    /// </summary>
    /// <param name="solution">The solution.</param>
    /// <param name="textControl">The text control.</param>
    public ContextAction(ISolution solution, ITextControl textControl) : base(solution, textControl) {
    }

    #endregion

    #region Public methods

    /// <summary>
    /// Determines whether the specified element is available.
    /// </summary>
    /// <param name="element">The element.</param>
    /// <returns>
    /// 	<c>true</c> if the specified element is available; otherwise, <c>false</c>.
    /// </returns>
    public static bool IsVisible(IElement element) {
      IInvocationExpression invocationExpression = element as IInvocationExpression;
      if(invocationExpression == null){
        return false;
      }

      IReference reference = invocationExpression.InvokedExpression as IReference;
      if(reference == null){
        return false;
      }

      ResolveResult resolveResult = reference.Resolve();
      if(resolveResult == null){
        return false;
      }

      IDeclaredElement declaredElement = resolveResult.DeclaredElement;
      if(declaredElement == null){
        return false;
      }

      XmlNode xmlNode = declaredElement.GetXMLDoc(true);
      if(xmlNode == null){
        return false;
      }

      XmlNodeList exceptionList = xmlNode.SelectNodes("exception");
      if(exceptionList == null || exceptionList.Count == 0){
        return false;
      }

      ITreeNode node = element as ITreeNode;
      if(node == null){
        return false;
      }

      List<string> exceptions = GetExceptions(exceptionList);

      while(node != null){
        ITryStatement tryStatement = node as ITryStatement;

        if(tryStatement != null){
          ExamineCatches(tryStatement, exceptions);

          if (exceptions.Count == 0){
            return false;
          }
        }

        node = node.Parent;
      }

      return exceptions.Count > 0;
    }

    #endregion

    #region Protected methods

    /// <summary>
    /// Executes the action.
    /// </summary>
    /// <param name="element">The element.</param>
    protected override void Execute(IElement element) {
      ITreeNode node = element.ToTreeNode();
      if(node == null){
        return;
      }

      IInvocationExpression invocationExpression = null;

      while(node != null){
        invocationExpression = node as IInvocationExpression;
        if(invocationExpression != null){
          break;
        }

        node = node.Parent;
      }

      if(invocationExpression == null){
        return;
      }

      Modify(delegate { CatchExceptions(invocationExpression); });
    }

    /// <summary>
    /// Check if this action is available at the constructed context.
    /// </summary>
    /// <param name="element">The element.</param>
    /// <returns>
    /// 	<c>true</c> if available; otherwise, <c>false</c>.
    /// </returns>
    protected override bool IsAvailable(IElement element) {
      ITreeNode node = element.ToTreeNode();
      if(node == null){
        return false;
      }

      IInvocationExpression invocationExpression = null;

      while(node != null){
        invocationExpression = node as IInvocationExpression;

        if(invocationExpression != null){
          break;
        }

        node = node.Parent;
      }

      if(invocationExpression == null){
        return false;
      }

      return IsVisible(invocationExpression);
    }

    /// <summary>
    /// Gets the text.
    /// </summary>
    /// <returns>The text.</returns>
    protected override string GetText() {
      return "Catch exceptions";
    }

    #endregion

    #region Private methods

    /// <summary>
    /// Generates the function assert statements.
    /// </summary>
    /// <param name="element">The element.</param>
    void CatchExceptions(IElement element) {
      IInvocationExpression invocationExpression = element as IInvocationExpression;
      if(invocationExpression == null){
        return;
      }

      IStatement statement = invocationExpression.GetContainingStatement();
      if(statement == null){
        return;
      }

      CSharpElementFactory factory = CSharpElementFactory.GetInstance(Solution);
      if(factory == null){
        return;
      }

      IReference reference = invocationExpression.InvokedExpression as IReference;
      if(reference == null){
        return;
      }

      ResolveResult resolveResult = reference.Resolve();
      if(resolveResult == null){
        return;
      }

      IDeclaredElement declaredElement = resolveResult.DeclaredElement;
      if(declaredElement == null){
        return;
      }

      XmlNode xmlNode = declaredElement.GetXMLDoc(true);
      if(xmlNode == null){
        return;
      }

      XmlNodeList exceptionList = xmlNode.SelectNodes("exception");
      if(exceptionList == null || exceptionList.Count == 0){
        return;
      }

      List<Pair<string, Type>> exceptions = GetSortedExceptions(exceptionList);

      string code = "try { " + statement.GetText() + " } ";

      foreach(Pair<string, Type> exception in exceptions) {
        code += "catch(" + exception.Key + ") { } ";
      }

      ITryStatement tryStatement = factory.CreateStatement(code) as ITryStatement;

      if(tryStatement == null){
        MessageBox.Show("Failed to create code.", "ReSharper");
        return;
      }

      statement.ReplaceBy(tryStatement);
    }

    /// <summary>
    /// Gets the exceptions.
    /// </summary>
    /// <param name="exceptionList">The exception list.</param>
    /// <returns>The exceptions.</returns>
    List<Pair<string, Type>> GetSortedExceptions(XmlNodeList exceptionList) {
      List<Pair<string, Type>> exceptions = new List<Pair<string, Type>>(exceptionList.Count);

      foreach(XmlNode exceptionNode in exceptionList){
        XmlAttribute attribute = exceptionNode.Attributes["cref"];
        if(attribute == null){
          continue;
        }

        string typeName = attribute.Value;
        if(string.IsNullOrEmpty(typeName)){
          continue;
        }

        if(typeName.StartsWith("T:")) {
          typeName = typeName.Substring(2);
        }

        Type exceptionType = Type.GetType(typeName, false, false);

        Pair<string, Type> pair = new Pair<string, Type>();

        pair.Key = typeName;
        pair.Value = exceptionType;

        exceptions.Add(pair);
      }

      exceptions.Sort(this);

      return exceptions;
    }

    /// <summary>
    /// Examines the catches.
    /// </summary>
    /// <param name="tryStatement">The try statement.</param>
    /// <param name="exceptions">The exceptions.</param>
    /// <returns>The catches.</returns>
    static void ExamineCatches(ITryStatement tryStatement, List<string> exceptions) {
      IList<ICatchClause> list = tryStatement.Catches;
      List<string> catches = new List<string>();

      foreach(ICatchClause clause in list){
        IDeclaredType declaredType = clause.ExceptionType;

        if(declaredType == null){
          break;
        }

        string clrName = declaredType.GetCLRName();

        if(!string.IsNullOrEmpty(clrName)){
          catches.Add(clrName);
        }
      }

      for(int n = exceptions.Count - 1; n >= 0; n--){
        string type = exceptions[n];

        if(catches.Contains(type)){
          exceptions.RemoveAt(n);
        }
      }
    }

    /// <summary>
    /// Gets the exceptions.
    /// </summary>
    /// <param name="exceptionList">The exception list.</param>
    /// <returns>The exceptions.</returns>
    static List<string> GetExceptions(XmlNodeList exceptionList) {
      List<string> result = new List<string>();

      foreach(XmlNode exceptionNode in exceptionList){
        XmlAttribute attribute = exceptionNode.Attributes["cref"];

        if(attribute == null){
          continue;
        }

        string typeName = attribute.Value;

        if(!string.IsNullOrEmpty(typeName)){
          if(typeName.StartsWith("T:")){
            typeName = typeName.Substring(2);
          }

          result.Add(typeName);
        }
      }

      return result;
    }

    #endregion

    #region IComparer<Pair<string,Type>> Members

    ///<summary>
    ///Compares two objects and returns a value indicating whether one is less than, equal to, or greater than the other.
    ///</summary>
    ///
    ///<returns>
    ///Value Condition Less than zerox is less than y.Zerox equals y.Greater than zerox is greater than y.
    ///</returns>
    ///
    ///<param name="y">The second object to compare.</param>
    ///<param name="x">The first object to compare.</param>
    int IComparer<Pair<string, Type>>.Compare(Pair<string, Type> x, Pair<string, Type> y) {
      if (x.Value == null){
        return -1;
      }

      if(y.Value == null) {
        return 1;
      }

      if(x.Value.IsSubclassOf(y.Value)) {
        return -1;
      }

      if(y.Value.IsSubclassOf(x.Value)) {
        return 1;
      }

      return string.Compare(x.Key, y.Key);
    }

    #endregion
  }
}
  