using System.Collections.Generic;
using System.Collections.Specialized;
using JetBrains.ReSharper.Daemon;
using JetBrains.ReSharper.ProjectModel;
using JetBrains.ReSharper.Psi;
using JetBrains.ReSharper.Psi.CSharp;
using JetBrains.ReSharper.Psi.CSharp.Tree;
using JetBrains.ReSharper.Psi.Tree;
using JetBrains.ReSharper.TextControl;
using ReSharper.Framework;

namespace ReSharper.ParameterAsserts {
  /// <summary>
  /// Implements a Generate Assertions powertoy plugin.
  /// </summary>
  [ContextAction(Name = "ParameterAsserts")]
  public class ContextAction : ContextActionBase {
    #region Constructor

    /// <summary>
    /// Initializes a new instance of the <see cref="ContextAction"/> class.
    /// </summary>
    /// <param name="solution">The solution.</param>
    /// <param name="textControl">The text control.</param>
    public ContextAction(ISolution solution, ITextControl textControl) : base(solution, textControl) {
    }

    #endregion

    #region Protected methods

    /// <summary>
    /// Executes this instance.
    /// </summary>
    /// <param name="element">The element.</param>
    protected override void Execute(IElement element) {
      IFunctionDeclaration functionDeclaration = GetFunctionDeclaration(element);

      if(functionDeclaration != null) {
        Modify(delegate { GenerateFunctionAsserts(functionDeclaration); });
        return;
      }

      functionDeclaration = GetContainingFunctionDeclaration(element);
      
      IParameterDeclaration parameterDeclaration = GetContainingParameterDeclaration(element);

      Modify(delegate { GenerateParameterAssert(functionDeclaration, parameterDeclaration); });
    }

    /// <summary>
    /// Determines whether this instance is available.
    /// </summary>
    /// <param name="element">The element.</param>
    /// <returns>
    /// 	<c>true</c> if this instance is available; otherwise, <c>false</c>.
    /// </returns>
    protected override bool IsAvailable(IElement element) {
      if(IsFunctionAvailable(element)) {
        return true;
      }

      if(IsParameterAvailable(element)) {
        return true;
      }

      return false;
    }

    /// <summary>
    /// Gets the text.
    /// </summary>
    /// <returns>The text.</returns>
    protected override string GetText() {
      return "Generate parameter assert(s)";
    }

    #endregion    

    #region Private methods

    /// <summary>
    /// Generates the function assert statements.
    /// </summary>
    /// <param name="functionDeclaration">The function declaration.</param>
    void GenerateFunctionAsserts(IFunctionDeclaration functionDeclaration) {
      CSharpElementFactory factory = CSharpElementFactory.GetInstance(Solution);

      IBlock body = functionDeclaration.Body;
      if(body == null){
        return;
      }

      IFunction function = functionDeclaration as IFunction;
      if(function == null){
        return;
      }

      List<IStatement> assertions = new List<IStatement>();

      foreach(IParameter parameter in function.Parameters) {
        if(!ParameterAssert.RequiresAssertion(parameter)) {
          continue;
        }

        IStatement statement;

        bool isAsserted = ParameterAssert.IsParameterAsserted(functionDeclaration, parameter, out statement);

        if (isAsserted){
          if(statement != null){
            body.RemoveStatement(statement);
            assertions.Add(factory.CreateStatement(statement.GetText()));
          }
        }
        else{
          string assertStatement = GetAssertStatement(parameter);
          if(assertStatement == null) {
            continue;
          }

          string statementText = string.Format(assertStatement, parameter.ShortName);

          statement = factory.CreateStatement(statementText);
          if(statement == null) {
            continue;
          }

          assertions.Add(statement);
        }
      }

      IStatement beforeStatement = null;

      IList<IStatement> list = body.Statements;
      if(list != null && list.Count > 0){
        beforeStatement = list[0];
      }

      for(int n = 0; n < assertions.Count; n++){
        body.AddStatementBefore(assertions[n], beforeStatement);
      }
    }

    /// <summary>
    /// Generates the parameter assert statement.
    /// </summary>
    /// <param name="functionDeclaration">The function declaration.</param>
    /// <param name="parameterDeclaration">The parameter declaration.</param>
    void GenerateParameterAssert(IFunctionDeclaration functionDeclaration, IParameterDeclaration parameterDeclaration) {
      CSharpElementFactory factory = CSharpElementFactory.GetInstance(Solution);

      IBlock body = functionDeclaration.Body;
      if(body == null){
        return;
      }

      IParameter parameter = parameterDeclaration as IParameter;
      if(parameter == null){
        return;
      }

      IStatement beforeStatement = null;

      IList<IStatement> statements = body.Statements;
      if(statements.Count > 0){
        beforeStatement = statements[0];
      }

      string assertStatement = GetAssertStatement(parameter);
      if(assertStatement == null) {
        return;
      }

      string statement = string.Format(assertStatement, parameter.ShortName);

      body.AddStatementBefore(factory.CreateStatement(statement), beforeStatement);
    }

    /// <summary>
    /// Gets the assert statement.
    /// </summary>
    /// <param name="parameter">The parameter.</param>
    /// <returns>The assert statement.</returns>
    static string GetAssertStatement(IParameter parameter) {
      IType type = parameter.Type;

      string typeName = type.GetLongPresentableName(parameter.Language);

      NameValueCollection typeAssertions = Settings.Instance.TypeAssertions;

      string result = typeAssertions[typeName];

      if(string.IsNullOrEmpty(result)){
        result = typeAssertions["*"];
      }

      return result;
    }

    /// <summary>
    /// Gets the containing function declaration.
    /// </summary>
    /// <param name="element">The element.</param>
    /// <returns>The containing function declaration.</returns>
    static IFunctionDeclaration GetContainingFunctionDeclaration(IElement element) {
      return element.GetContainingElement<IFunctionDeclaration>(false);
    }

    /// <summary>
    /// Gets the containing parameter declaration.
    /// </summary>
    /// <param name="element">The element.</param>
    /// <returns>The containing parameter declaration.</returns>
    static IParameterDeclaration GetContainingParameterDeclaration(IElement element) {
      return element.GetContainingElement<IParameterDeclaration>(false);
    }

    /// <summary>
    /// Gets the function declaration.
    /// </summary>
    /// <param name="element">The element.</param>
    /// <returns>The function declaration.</returns>
    static IFunctionDeclaration GetFunctionDeclaration(IElement element) {
      IIdentifierNode identifierNode = element as IIdentifierNode;
      if(identifierNode == null){
        return null;
      }

      return identifierNode.Parent as IFunctionDeclaration;
    }

    /// <summary>
    /// Determines whether [is function available] [the specified element].
    /// </summary>
    /// <param name="element">The element.</param>
    /// <returns>
    /// 	<c>true</c> if [is function available] [the specified element]; otherwise, <c>false</c>.
    /// </returns>
    static bool IsFunctionAvailable(IElement element) {
      IFunctionDeclaration functionDeclaration = GetFunctionDeclaration(element);

      if(functionDeclaration == null) {
        return false;
      }

      return ParameterAssert.IsFunctionMissingAssertions(functionDeclaration);
    }

    /// <summary>
    /// Determines whether [is parameter available] [the specified element].
    /// </summary>
    /// <param name="element">The element.</param>
    /// <returns>
    /// 	<c>true</c> if [is parameter available] [the specified element]; otherwise, <c>false</c>.
    /// </returns>
    static bool IsParameterAvailable(IElement element) {
      IParameterDeclaration parameterDeclaration = GetContainingParameterDeclaration(element);
      if(parameterDeclaration == null) {
        return false;
      }

      IFunctionDeclaration functionDeclaration = parameterDeclaration.GetContainingTypeMemberDeclaration() as IFunctionDeclaration;
      if(functionDeclaration == null) {
        return false;
      }

      return ParameterAssert.IsParameterMissingAssertion(functionDeclaration, parameterDeclaration);
    }

    #endregion
  }
}
  