using JetBrains.ReSharper.Psi;
using JetBrains.ReSharper.Psi.CSharp.Tree;
using JetBrains.ReSharper.Psi.Resolve;
using JetBrains.ReSharper.Psi.Tree;

namespace ReSharper.ParameterAsserts {
  /// <summary>
  /// Represents a FindParameterElementProcessor.
  /// </summary>
  public class FindParameterElementProcessor: IRecursiveElementProcessor {
    #region Fields

    readonly IParameter _parameter;
    bool _found;

    #endregion

    #region Constructor

    /// <summary>
    /// Initializes a new instance of the <see cref="FindParameterElementProcessor"/> class.
    /// </summary>
    /// <param name="parameter">The parameter.</param>
    public FindParameterElementProcessor(IParameter parameter) {
      _parameter = parameter;
    }

    #endregion

    #region Public properties

    /// <summary>
    /// Gets a value indicating whether the parameter was found.
    /// </summary>
    /// <value><c>true</c> if found; otherwise, <c>false</c>.</value>
    public bool Found {
      get {
        return _found;
      }
    }

    #endregion

    #region IRecursiveElementProcessor Members

    /// <summary>
    /// Determines if the interior should be processed.
    /// </summary>
    /// <param name="element">The element.</param>
    /// <returns><c>true</c> if the interior should be processed; otherwise, <c>false</c>.</returns>
    public bool InteriorShouldBeProcessed(IElement element) {
      return true;
    }

    /// <summary>
    /// Processes before the interior.
    /// </summary>
    /// <param name="element">The element.</param>
    public void ProcessBeforeInterior(IElement element) {
    }

    /// <summary>
    /// Processes after the interior.
    /// </summary>
    /// <param name="element">The element.</param>
    public void ProcessAfterInterior(IElement element) {
      if (_found){
        return;
      }

      IReferenceExpression referenceExpression = element as IReferenceExpression;
      if(referenceExpression == null) {
        return;
      }

      IReference reference = referenceExpression.Reference;
      if(reference == null) {
        return;
      }

      ResolveResult resolveResult = reference.Resolve();
      if(resolveResult == null) {
        return;
      }
      
      IParameter parameter = resolveResult.DeclaredElement as IParameter;
      if(parameter == null){
        return;
      }

      if (parameter.IsValueVariable && _parameter.IsValueVariable){
        _found = true;
      }
      else if(parameter == _parameter) {
        _found = true;
      }
    }

    /// <summary>
    /// Gets a value indicating whether the <see cref="FindParameterElementProcessor"/> processings is finished.
    /// </summary>
    /// <value>
    /// 	<c>true</c> if the <see cref="FindParameterElementProcessor"/> processings is finished; otherwise, <c>false</c>.
    /// </value>
    public bool ProcessingIsFinished {
      get {
        return _found;
      }
    }

    #endregion
  }
}