using System.Collections.Generic;
using System.Collections.Specialized;
using JetBrains.ReSharper.Psi;
using JetBrains.ReSharper.Psi.CSharp.Impl;
using JetBrains.ReSharper.Psi.CSharp.Tree;
using JetBrains.ReSharper.Psi.Resolve;

namespace ReSharper.ParameterAsserts {
  /// <summary>
  /// Implements support functions for working with assertions.
  /// </summary>
  public static class ParameterAssert {
    #region Public methods

    /// <summary>
    /// Determines whether the specified parameter is asserted.
    /// </summary>
    /// <param name="functionDeclaration">The function declaration.</param>
    /// <param name="parameter">The parameter.</param>
    /// <param name="statement">The statement.</param>
    /// <returns>
    /// 	<c>true</c> if the specified parameter is asserted; otherwise, <c>false</c>.
    /// </returns>
    public static bool IsParameterAsserted(IFunctionDeclaration functionDeclaration, IParameter parameter, out IStatement statement) {
      statement = null;

      IBlock block = functionDeclaration.Body;
      if(block == null){
        return true;
      }

      IList<IStatement> list = block.Statements;
      if(list == null){
        return false;
      }

      if(Settings.Instance.OmitBaseConstructors){
        IConstructorDeclaration constructorDeclaration = functionDeclaration as IConstructorDeclaration;

        if(constructorDeclaration != null){
          IConstructorInitializer initializer = constructorDeclaration.Initializer;

          if(initializer != null){
            if(IsParameterInvocationArgument(initializer.Arguments, parameter)){
              return true;
            }
          }
        }
      }

      foreach(IStatement stm in list) {
        IExpressionStatement expressionStatement = stm as IExpressionStatement;
        if(expressionStatement == null) {
          continue;
        }

        IInvocationExpression invocationExpression = expressionStatement.Expression as IInvocationExpression;
        if(invocationExpression == null) {
          continue;
        }

        bool isBaseInvocation = false;

        if (Settings.Instance.OmitBaseMethods){
          isBaseInvocation = IsBaseInvocation(invocationExpression);
        }

        if(isBaseInvocation || IsAssertInvocation(invocationExpression)) {
          if(IsParameterInvocationArgument(invocationExpression, parameter)) {
            statement = stm;
            return true;
          }
        }
      }

      return false;
    }

    /// <summary>
    /// Determines if the specified parameter requires assertion.
    /// </summary>
    /// <param name="parameter">The parameter.</param>
    /// <returns>
    /// 	<c>true</c> if the specified parameter requires asserted; otherwise, <c>false</c>.
    /// </returns>
    public static bool RequiresAssertion(IParameter parameter) {
      if(parameter.Kind == ParameterKind.OUTPUT){
        return false;
      }

      IInterface interfaceType = parameter.GetContainingType() as IInterface;
      if(interfaceType != null){
        return false;
      }

      IFunction function = parameter.GetContainingTypeMember() as IFunction;
      if(function == null){
        return false;
      }

      IOperator @operator = function as IOperator;
      if(@operator != null){
        return false;
      }

      IConversionOperator conversionOperator = function as IConversionOperator;
      if(conversionOperator != null){
        return false;
      }

      if(function.ShortName == "Equals"){
        return false;
      }

      Settings settings = Settings.Instance;

      AccessRights rights = function.GetAccessRights();

      if (!settings.PublicMethods && rights == AccessRights.PUBLIC){
        return false;
      }

      if(!settings.InternalMethods && rights == AccessRights.INTERNAL) {
        return false;
      }

      if(!settings.ProtectedMethods && rights == AccessRights.PROTECTED) {
        return false;
      }

      if(!settings.PrivateMethods && rights == AccessRights.PRIVATE) {
        return false;
      }

      return !CSharpTypeConversionRule.Instance.IsValueType(parameter.Type);
    }

    /// <summary>
    /// Determines whether if the specified function is missing assertions.
    /// </summary>
    /// <param name="functionDeclaration">The function declaration.</param>
    /// <returns>
    /// 	<c>true</c> if the specified function is missing assertions; otherwise, <c>false</c>.
    /// </returns>
    public static bool IsFunctionMissingAssertions(IFunctionDeclaration functionDeclaration) {
      IFunction function = functionDeclaration as IFunction;
      if (function == null){
        return false;
      }

      if(function.Parameters.Length <= 0 || function.IsAbstract || function.IsExtern){
        return false;
      }

      foreach(IParameter parameter in function.Parameters){
        if(parameter == null){
          continue;
        }

        IStatement statement;

        if(RequiresAssertion(parameter) && !IsParameterAsserted(functionDeclaration, parameter, out statement)){
          return true;
        }
      }

      return false;
    }

    /// <summary>
    /// Determines whether the specified parameter is missing an assertion.
    /// </summary>
    /// <param name="functionDeclaration">The function declaration.</param>
    /// <param name="parameterDeclaration">The parameter declaration.</param>
    /// <returns>
    /// 	<c>true</c> if the specified parameter is missing an assertion; otherwise, <c>false</c>.
    /// </returns>
    public static bool IsParameterMissingAssertion(IFunctionDeclaration functionDeclaration, IParameterDeclaration parameterDeclaration) {
      IParameter parameter = parameterDeclaration as IParameter;

      if(parameter == null){
        return false;
      }

      IStatement statement;

      if(RequiresAssertion(parameter) && !IsParameterAsserted(functionDeclaration, parameter, out statement)) {
        return true;
      }

      return false;
    }

    #endregion

    #region Private methods

    /// <summary>
    /// Gets the full name of the method.
    /// </summary>
    /// <param name="method">The method.</param>
    /// <returns>The fully qualified name of the method.</returns>
    static string GetFullMethodName(IMethod method) {
      string result = method.ShortName;

      ITypeElement typeElement = method.GetContainingType();
      if(typeElement != null) {
        result = typeElement.ShortName + "." + result;

        INamespace ns = typeElement.GetContainingNamespace();
        if(ns != null) {
          result = ns.QualifiedName + "." + result;
        }
      }

      return result;
    }

    /// <summary>
    /// Determines whether the specified expression is an assertion invocation.
    /// </summary>
    /// <param name="invocationExpression">The invocation expression.</param>
    /// <returns>
    /// 	<c>true</c> if the specified expression is an assertion invocation; otherwise, <c>false</c>.
    /// </returns>
    public static bool IsAssertInvocation(IInvocationExpression invocationExpression) {
      IReference reference = invocationExpression.InvokedExpression as IReference;
      if(reference == null) {
        return false;
      }

      ResolveResult resolveResult = reference.Resolve();
      if(resolveResult == null) {
        return false;
      }

      IMethod method = null;

      IMethodDeclaration methodDeclaration = resolveResult.DeclaredElement as IMethodDeclaration;
      if(methodDeclaration != null) {
        method = methodDeclaration as IMethod;
      }

      if(method == null) {
        method = resolveResult.DeclaredElement as IMethod;
      }

      if(method == null) {
        return false;
      }

      string methodName = GetFullMethodName(method);

      StringCollection methods = Settings.Instance.Methods;

      foreach(string assertMethods in methods) {
        if(methodName == assertMethods) {
          return true;
        }
      }

      return false;
    }

    /// <summary>
    /// Determines whether the specified invocation expression is a base invocation.
    /// </summary>
    /// <param name="invocationExpression">The invocation expression.</param>
    /// <returns>
    /// 	<c>true</c> if the specified invocation expression is a base invocation; otherwise, <c>false</c>.
    /// </returns>
    static bool IsBaseInvocation(IInvocationExpression invocationExpression) {
      IReferenceExpression reference = invocationExpression.InvokedExpression as IReferenceExpression;
      if(reference == null) {
        return false;
      }

      return reference.QualifierExpression is IBaseExpression;
    }

    /// <summary>
    /// Determines whether [is parameter invocation argument] [the specified invocation expression].
    /// </summary>
    /// <param name="invocationExpression">The invocation expression.</param>
    /// <param name="parameter">The parameter.</param>
    /// <returns>
    /// 	<c>true</c> if [is parameter invocation argument] [the specified invocation expression]; otherwise, <c>false</c>.
    /// </returns>
    static bool IsParameterInvocationArgument(IInvocationExpression invocationExpression, IParameter parameter) {
      return IsParameterInvocationArgument(invocationExpression.Arguments, parameter);
    }

    /// <summary>
    /// Determines if the specified parameter is an argument in the invocation expression.
    /// </summary>
    /// <param name="arguments">The arguments.</param>
    /// <param name="parameter">The parameter.</param>
    /// <returns>
    /// 	<c>true</c> if the specified parameter is an argument in the invocation expression; otherwise, <c>false</c>.
    /// </returns>
    static bool IsParameterInvocationArgument(IList<IArgument> arguments, IParameter parameter) {
      foreach(IArgument argument in arguments) {
        FindParameterElementProcessor findParameterElementProcessor = new FindParameterElementProcessor(parameter);

        argument.ProcessDescendants(findParameterElementProcessor);

        if(findParameterElementProcessor.Found) {
          return true;
        }
      }

      return false;
    }

    #endregion
  }
}
