using System.Collections.Specialized;
using System.Text;
using System.Xml;
using JetBrains.ReSharper.Component;
using JetBrains.ReSharper.Shell;
using JetBrains.ReSharper.Util;

namespace ReSharper.ParameterAsserts {
  [ShellComponentInterface(ProgramConfigurations.VS_ADDIN)]
  [ShellComponentImplementation()]
  public class Settings : IShellComponent, IXmlExternalizableShellComponent {
    #region Fields

    StringCollection _methods;
    NameValueCollection _typeAssertions;
    bool _publicMethods = true;
    bool _internalMethods = true;
    bool _protectedMethods = true;
    bool _privateMethods = true;
    bool _omitBaseConstructors = true;
    bool _omitBaseMethods = false;

    #endregion

    #region Public properties

    /// <summary>
    /// Gets the instance.
    /// </summary>
    /// <value>The instance.</value>
    public static Settings Instance {
      get {
        return (Settings)Shell.Instance.GetComponent(typeof(Settings));
      }
    }

    /// <summary>
    /// Gets or sets a value indicating whether the <see cref="Settings"/> omits the base constructors.
    /// </summary>
    /// <value>
    /// 	<c>true</c> if the <see cref="Settings"/> omits the base constructors; otherwise, <c>false</c>.
    /// </value>
    [XmlExternalizationUtil.ExternalizableAttribute(DefaultValue = true)]
    public bool OmitBaseConstructors {
      get {
        return _omitBaseConstructors;
      }
      set {
        _omitBaseConstructors = value;
      }
    }

    /// <summary>
    /// Gets or sets a value indicating whether the <see cref="Settings"/> omits the base methods.
    /// </summary>
    /// <value>
    /// 	<c>true</c> if the <see cref="Settings"/> omits the base methods; otherwise, <c>false</c>.
    /// </value>
    [XmlExternalizationUtil.ExternalizableAttribute(DefaultValue = true)]
    public bool OmitBaseMethods {
      get {
        return _omitBaseMethods;
      }
      set {
        _omitBaseMethods = value;
      }
    }

    /// <summary>
    /// Gets or sets a value indicating whether the <see cref="Settings"/> publics the methods.
    /// </summary>
    /// <value>
    /// 	<c>true</c> if the <see cref="Settings"/> publics the methods; otherwise, <c>false</c>.
    /// </value>
    [XmlExternalizationUtil.ExternalizableAttribute(DefaultValue = true)]
    public bool PublicMethods {
      get {
        return _publicMethods;
      }
      set {
        _publicMethods = value;
      }
    }

    /// <summary>
    /// Gets or sets a value indicating whether the <see cref="Settings"/> internals the methods.
    /// </summary>
    /// <value>
    /// 	<c>true</c> if the <see cref="Settings"/> internals the methods; otherwise, <c>false</c>.
    /// </value>
    [XmlExternalizationUtil.ExternalizableAttribute(DefaultValue = true)]
    public bool InternalMethods {
      get {
        return _internalMethods;
      }
      set {
        _internalMethods = value;
      }
    }

    /// <summary>
    /// Gets or sets a value indicating whether the <see cref="Settings"/> protecteds the methods.
    /// </summary>
    /// <value>
    /// 	<c>true</c> if the <see cref="Settings"/> protecteds the methods; otherwise, <c>false</c>.
    /// </value>
    [XmlExternalizationUtil.ExternalizableAttribute(DefaultValue = true)]
    public bool ProtectedMethods {
      get {
        return _protectedMethods;
      }
      set {
        _protectedMethods = value;
      }
    }

    /// <summary>
    /// Gets or sets a value indicating whether the <see cref="Settings"/> privates the methods.
    /// </summary>
    /// <value>
    /// 	<c>true</c> if the <see cref="Settings"/> privates the methods; otherwise, <c>false</c>.
    /// </value>
    [XmlExternalizationUtil.ExternalizableAttribute(DefaultValue = true)]
    public bool PrivateMethods {
      get {
        return _privateMethods;
      }
      set {
        _privateMethods = value;
      }
    }

    /// <summary>
    /// Gets or sets the methods.
    /// </summary>
    /// <value>The methods.</value>
    public StringCollection Methods {
      get {
        if(_methods == null){
          InitMethods();
        }

        return _methods;
      }
      set {
        _methods = value;
      }
    }
    /// <summary>
    /// Gets or sets the methods in a serializable format.
    /// </summary>                             
    /// <remarks>This is for serialization only.</remarks>
    /// <value>The methods.</value>
    [XmlExternalizationUtil.ExternalizableAttribute(DefaultValue = true)]
    public string SerializableMethods {
      get {
        StringBuilder result = new StringBuilder();

        bool first = true;

        foreach(string method in Methods){
          if(!first){
            result.Append("|");
          }

          first = false;

          result.Append(method);
        }

        return result.ToString();
      }
      set {
        _methods = new StringCollection();

        if(!string.IsNullOrEmpty(value)){
          _methods.AddRange(value.Split('|'));
        }
      }
    }

    /// <summary>
    /// Gets or sets the type assertions.
    /// </summary>
    /// <remarks>This is for serialization only.</remarks>
    /// <value>The type assertions.</value>
    [XmlExternalizationUtil.ExternalizableAttribute(DefaultValue = true)]
    public string SerializableTypeAssertions {
      get {
        StringBuilder result = new StringBuilder();

        bool first = true;

        foreach(string key in TypeAssertions.Keys) {
          if(!first) {
            result.Append("|");
          }

          first = false;

          result.Append(key);
          result.Append('^');
          result.Append(TypeAssertions[key]);
        }

        return result.ToString();
      }
      set {
        _typeAssertions = new NameValueCollection();  

        if(!string.IsNullOrEmpty(value)) {
          foreach(string pair in value.Split('|')){
            string[] parts = pair.Split('^');

            _typeAssertions.Add(parts[0], parts[1]);
          }
        }
      }
    }

    /// <summary>
    /// Gets or sets the type assertions.
    /// </summary>
    /// <value>The type assertions.</value>
    public NameValueCollection TypeAssertions {
      get {
        if(_typeAssertions == null){
          InitTypeAssertions();
        }

        return _typeAssertions;
      }
      set {
        _typeAssertions = value;
      }
    }

    #endregion

    #region IShellComponent implementation

    /// <summary>
    /// Inits this instance.
    /// </summary>
    public void Init() {
    }

    /// <summary>
    /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
    /// </summary>
    public void Dispose() {
    }

    #endregion

    #region IXmlExternalizableShellComponent implementation

    /// <summary>
    /// Gets the name of the tag.
    /// </summary>
    /// <value>The name of the tag.</value>
    public string TagName {
      get {
        return "ReSharper.PowerToys.GenerateAssertions";
      }
    }

    /// <summary>
    /// This method must not fail with null or unexpected Xml!!!
    /// </summary>
    /// <param name="element"></param>
    public void ReadFromXml(XmlElement element) {
      if (element == null){
        InitMethods();
        InitTypeAssertions();
        return;
      }

      XmlExternalizationUtil.ReadFromXml(element, this);
    }

    /// <summary>
    /// Writes to XML.
    /// </summary>
    /// <param name="element">The element.</param>
    /// <returns></returns>
    public bool WriteToXml(XmlElement element) {
      return XmlExternalizationUtil.WriteToXml(element, this);
    }

    /// <summary>
    /// Gets the externalization scope.
    /// </summary>
    /// <value>The externalization scope.</value>
    public ShellComponentExternalizationScope ExternalizationScope {
      get {
        return ShellComponentExternalizationScope.SETTINGS;
      }
    }

    #endregion

    #region Private methods

    /// <summary>
    /// Inits the methods.
    /// </summary>
    void InitMethods() {
      _methods = new StringCollection();

      _methods.Add("System.Diagnostics.Debug.Assert");
    }

    /// <summary>
    /// Inits the type assertions.
    /// </summary>
    void InitTypeAssertions() {
      _typeAssertions = new NameValueCollection();

      _typeAssertions.Add("*", "Debug.Assert({0} != null, \"Parameter '{0}' cannot be null.\");");
    }

    #endregion
  }
}
