using System;
using System.Collections.Generic;
using System.Reflection;
using JetBrains.ReSharper.Daemon;
using JetBrains.ReSharper.ProjectModel;
using JetBrains.ReSharper.Psi;
using JetBrains.ReSharper.Psi.Tree;
using JetBrains.ReSharper.Shell.Progress;
using JetBrains.ReSharper.Util;

namespace ReSharper.Framework {
  /// <summary>
  /// Represents a simple context action.
  /// </summary>
  public abstract class HighlighterBase: IDaemonStage, IDaemonStageProcess, IRecursiveElementProcessor {
    #region Fields

    readonly IDaemonProcess _daemonProcess;
    List<IHighlighting> _highlightings;

    #endregion

    #region Constructor

    /// <summary>
    /// Initializes a new instance of the <see cref="HighlighterBase"/> class.
    /// </summary>
    protected HighlighterBase() {
    }

    /// <summary>
    /// Initializes a new instance of the <see cref="HighlighterBase"/> class.
    /// </summary>
    /// <param name="daemonProcess">The daemon process.</param>
    public HighlighterBase(IDaemonProcess daemonProcess) {
      _daemonProcess = daemonProcess;
    }

    #endregion

    #region Public properties

    /// <summary>
    /// Gets the daemon process.
    /// </summary>
    /// <value>The daemon process.</value>
    public IDaemonProcess DaemonProcess {
      get {
        return _daemonProcess;
      }
    }

    /// <summary>
    /// Gets the highlightings.
    /// </summary>
    /// <value>The highlightings.</value>
    public List<IHighlighting> Highlightings {
      get {
        return _highlightings;
      }
    }

    #endregion

    #region Protected methods

    /// <summary>
    /// Adds the warning.
    /// </summary>
    /// <param name="element">The element.</param>
    /// <param name="tooltip">The message.</param>
    protected void AddHighlighting(IElement element, string tooltip) {
      HighlightingBase highlightingBase = new HighlightingBase(element, tooltip);

      AddHighlighting(highlightingBase);
    }

    /// <summary>
    /// Adds the warning.
    /// </summary>
    /// <param name="range">The range.</param>
    /// <param name="tooltip">The message.</param>
    protected void AddHighlighting(DocumentRange range, string tooltip) {
      HighlightingBase highlightingBase = new HighlightingBase(range, tooltip);

      AddHighlighting(highlightingBase);
    }

    /// <summary>
    /// Adds the warning.
    /// </summary>
    /// <param name="highlighting">The highlighting.</param>
    protected void AddHighlighting(IHighlighting highlighting) {
      Highlightings.Add(highlighting);
    }

    /// <summary>
    /// Check if the error stripe indicator is needed for this stage after processing given <paramref name="projectFile"/>
    /// </summary>
    /// <param name="projectFile"></param>
    /// <returns></returns>
    protected virtual ErrorStripeRequest NeedsErrorStripe(IProjectFile projectFile) {
      return ErrorStripeRequest.STRIPE_AND_ERRORS;
    }

    /// <summary>
    /// Processes the element.
    /// </summary>
    /// <param name="element">The element.</param>
    protected abstract void ProcessElement(IElement element);

    #endregion

    #region IDaemonStage Members

    /// <summary>
    /// Gets a value indicating whether [run for invisible documents].
    /// </summary>
    /// <value>
    /// 	<c>true</c> if [run for invisible documents]; otherwise, <c>false</c>.
    /// </value>
    bool IDaemonStage.RunForInvisibleDocuments {
      get {
        return false;
      }
    }
    /// <summary>
    /// Specifies stages that should be executed before this stage.
    /// </summary>
    Type[] IDaemonStage.StagesBefore {
      get {
        return new Type[] { typeof(LanguageSpecificDaemonStage) };
      }
    }
    /// <summary>
    /// Specifies stages that should be executed after this stage.
    /// </summary>
    Type[] IDaemonStage.StagesAfter {
      get {
        return EmptyArray<Type>.Instance;
      }
    }

    #region IRecursiveElementProcessor Members

    /// <summary>
    /// Interiors the should be processed.
    /// </summary>
    /// <param name="element">The element.</param>
    /// <returns>The should be processed.</returns>
    bool IRecursiveElementProcessor.InteriorShouldBeProcessed(IElement element) {
      return true;
    }

    /// <summary>
    /// Processes the before interior.
    /// </summary>
    /// <param name="element">The element.</param>
    void IRecursiveElementProcessor.ProcessBeforeInterior(IElement element) {
    }

    /// <summary>
    /// Processes the after interior.
    /// </summary>
    /// <param name="element">The element.</param>
    void IRecursiveElementProcessor.ProcessAfterInterior(IElement element) {
      ProcessElement(element);
    }

    /// <summary>
    /// Gets a value indicating whether [processing is finished].
    /// </summary>
    /// <value>
    /// 	<c>true</c> if [processing is finished]; otherwise, <c>false</c>.
    /// </value>
    bool IRecursiveElementProcessor.ProcessingIsFinished {
      get {
        return _daemonProcess.InterruptFlag;
      }
    }

    #endregion

    ///<summary>
    ///  Creates a code analysis process corresponding to this stage for analysing a file.
    ///</summary>
    ///<returns>
    ///Code analysis process to be executed or 
    ///<c>null</c> if this stage is not available for this file.
    ///</returns>
    IDaemonStageProcess IDaemonStage.CreateProcess(IDaemonProcess process) {
      if(process == null) {
        throw new ArgumentNullException("process");
      }

      Type type = GetType();

      ConstructorInfo constructor = type.GetConstructor(new Type[] { typeof(IDaemonProcess) });
      if(constructor == null){
        return null;
      }

      return constructor.Invoke(new object[] { process } ) as IDaemonStageProcess;
    }

    ///<summary>
    /// Check if the error stripe indicator is needed for this stage after processing given <paramref name="projectFile" />
    ///</summary>
    ErrorStripeRequest IDaemonStage.NeedsErrorStripe(IProjectFile projectFile) {
      return NeedsErrorStripe(projectFile);
    }

    #endregion

    #region IDaemonStageProcess Members

    ///<summary>
    /// Executes the process and returns resulting highlightings and embedded objects to be inserted into the editor.
    /// The process should check for <see cref="P:JetBrains.ReSharper.Daemon.IDaemonProcess.InterruptFlag" /> periodically (with intervals less than 100 ms)
    /// and throw <see cref="T:JetBrains.ReSharper.Shell.Progress.ProcessCancelledException" /> if it is true. 
    /// Failing to do so may cause the program to prevent user from typing while analysing the code.
    ///</summary>
    ///<returns>
    ///New highlightings and embedded objects. Return 
    ///<c>null</c> if this stage doesn't produce
    ///            any of them.
    ///</returns>
    DaemonStageProcessResult IDaemonStageProcess.Execute() {
      DaemonStageProcessResult result = new DaemonStageProcessResult();

      PsiManager manager = PsiManager.GetInstance(_daemonProcess.Solution);
      if(manager == null){
        return result;
      }
      
      IFile file = manager.GetPsiFile(_daemonProcess.ProjectFile);
      if(file == null) {
        return result;
      }

      _highlightings = new List<IHighlighting>();

      file.ProcessDescendants(this);

      if(_daemonProcess.InterruptFlag) {
        throw new ProcessCancelledException();
      }

      result.FullyRehighlighted = true;
      result.Highlightings = _highlightings.ToArray();

      return result;
    }

    #endregion
  }
}
