using System.Drawing;
using JetBrains.ReSharper.Daemon;
using JetBrains.ReSharper.Psi;
using JetBrains.ReSharper.Psi.Tree;

namespace ReSharper.Framework {
  /// <summary>
  /// The highlighting that warns about high complexity
  /// </summary>
  public class HighlightingBase : IHighlighting {
    #region Fields

    readonly DocumentRange _range;
    string _tooltip;
    string _attributeId = HighlightingAttributeIds.WARNING_ATTRIBUTE;
    OverlapResolvePolicy _overlapResolvePolicy = OverlapResolvePolicy.WARNING;
    Severity _severity = Severity.WARNING;
    Color _colorOnStripe = Color.Olive;
    int _navigationOffsetPatch;

    #endregion

    #region Constructor

    /// <summary>
    /// Initializes a new instance of the <see cref="HighlightingBase"/> class.
    /// </summary>
    /// <param name="element">The element.</param>
    /// <param name="toolTip">The tool tip.</param>
    public HighlightingBase(IElement element, string toolTip) {
      _range = element.GetDocumentRange();
      _tooltip = toolTip;
    }

    /// <summary>
    /// Initializes a new instance of the <see cref="HighlightingBase"/> class.
    /// </summary>
    /// <param name="range">The range.</param>
    /// <param name="toolTip">The tool tip.</param>
    public HighlightingBase(DocumentRange range, string toolTip) {
      _range = range;
      _tooltip = toolTip;
    }

    #endregion

    #region IHighlighting Members

    /// <summary>
    /// Get the original range of the highlighting.
    /// NOTE: This range is valid only when the error is created. 
    /// After the highlighting is registered in the markup model, the corresponding 
    /// range marker should be used to acquire the range
    /// </summary>
    DocumentRange IHighlighting.Range {
      get {
        return _range;
      }
    }
    /// <summary>
    /// Attribute of this highlighting in the markup model
    /// </summary>
    public string AttributeId {
      get {
        return _attributeId;
      }
      set {
        _attributeId = value;
      }
    }

    /// <summary>
    /// Get the ovelap resolving policy for this highlighting
    /// </summary>
    public OverlapResolvePolicy OverlapResolvePolicy {
      get {
        return _overlapResolvePolicy;
      }
      set {
        _overlapResolvePolicy = value;
      }
    }
    /// <summary>
    /// Get the severity of this highlighting
    /// </summary>
    public Severity Severity {
      get {
        return _severity;
      }
      set {
        _severity = value;
      }
    }

    /// <summary>
    /// Tooltip for this highlighting
    /// </summary>
    public string ToolTip {
      get {
        return _tooltip;
      }
      set {
        _tooltip = value;
      }
    }

    /// <summary>
    /// Color on gutter for this highlighting
    /// NOTE: Will be called only if Severity == INFO
    /// </summary>
    public Color ColorOnStripe {
      get {
        return _colorOnStripe; 
      }
      set {
        _colorOnStripe = value;
      }
    }

    /// <summary>
    /// Specifies the offset from the Range.StartOffset to set the cursor to when navigating 
    /// to this highlighting. Usually returns 
    /// <c>0</c>
    /// </summary>
    public int NavigationOffsetPatch {
      get {
        return _navigationOffsetPatch;
      }
      set {
        _navigationOffsetPatch = value;
      }
    }

    #endregion
  }
}
