using System;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using JetBrains.ReSharper.Daemon;
using JetBrains.ReSharper.Editor;
using JetBrains.ReSharper.ProjectModel;
using JetBrains.ReSharper.Psi;
using JetBrains.ReSharper.Psi.CSharp.Tree;
using JetBrains.ReSharper.Psi.Tree;
using JetBrains.ReSharper.Shell;
using JetBrains.ReSharper.TextControl;
using JetBrains.ReSharper.Util;

namespace ReSharper.SynchronizeNamespaceWithFilePath {
  /// <summary>
  /// Implements a Generate Assertions powertoy plugin.
  /// </summary>
  public class ChangeNamespace : IBulbItem {
    #region Fields

    readonly ISolution _solution;
    readonly ITextControl _textControl;

    #endregion

    #region Constructor

    /// <summary>
    /// Initializes a new instance of the <see cref="ContextAction"/> class.
    /// </summary>
    /// <param name="solution">The solution.</param>
    /// <param name="textControl">The text control.</param>
    public ChangeNamespace(ISolution solution, ITextControl textControl) {
      _solution = solution;
      _textControl = textControl;
    }

    #endregion

    #region Public properties

    /// <summary>
    /// Gets the text.
    /// </summary>
    /// <value>The text.</value>
    public string Text {
      get {
        IProjectFile projectFile = DocumentManager.GetInstance(_solution).GetProjectFile(_textControl.Document);
        if(projectFile == null) {
          return "Change namespace";
        }

        return "Change namespace to \"" + GetNamespace(projectFile) + "\"";
      }
    }

    #endregion

    #region Public methods

    /// <summary>
    /// Executes the action.
    /// </summary>
    /// <param name="solution">The solution.</param>
    /// <param name="textControl">The text control.</param>
    public void Execute(ISolution solution, ITextControl textControl) {
      if(_solution != solution || _textControl != textControl){
        throw new InvalidOperationException();
      }

      using(ModificationCookie cookie = _textControl.Document.EnsureWritable()) {
        if(cookie.EnsureWritableResult != EnsureWritableResult.SUCCESS) {
          return;
        }

        using(CommandCookie.Create(string.Format("Context Action {0}", Text))) {
          PsiManager.GetInstance(_solution).DoTransaction(delegate {
            ChangeNamespaceName();
          });
        }
      }
    }

    /// <summary>
    /// Gets the target file.
    /// </summary>
    /// <param name="projectFile">The project file.</param>
    /// <returns>The target file.</returns>
    public string GetNamespace(IProjectFile projectFile) {
      IProject project = projectFile.GetProject();
      if(project == null) {
        return null;
      }

      FileSystemPath folder = projectFile.Location;

      folder = folder.ConvertToRelativePath(project.Location);

      string path = Path.GetDirectoryName(folder.ToString());

      path = path.Replace("\\", ".");

      path = ConvertToPascalCase(path);

      path = Regex.Replace(path, "[^0-9a-zA-Z\\.]", string.Empty);

      return path;
    }

    /// <summary>
    /// Converts to pascal case.
    /// </summary>
    /// <param name="path">The identifier.</param>
    /// <returns>The to pascal case.</returns>
    static string ConvertToPascalCase(string path) {
      StringBuilder s = new StringBuilder(path);

      int first = -1;

      for(int i = s.Length - 2; i >= 0; i--) {
        if(!char.IsLetterOrDigit(s[i])) {
          s[i + 1] = Char.ToUpper(s[i + 1]);
        }
        else {
          first = i;
        }
      }

      if(first >= 0) {
        s[first] = Char.ToUpper(s[first]);
      }

      return s.ToString();
    }


    #endregion

    #region Private methods

    /// <summary>
    /// Changes the name of the namespace.
    /// </summary>
    void ChangeNamespaceName() {
      IElement element = GetElementAtCaret();
      if(element == null) {
        return;
      }

      IProjectFile projectFile = DocumentManager.GetInstance(_solution).GetProjectFile(_textControl.Document);
      if(projectFile == null) {
        return;
      }

      INamespaceDeclaration namespaceDeclaration = ContextAction.GetNamespaceDeclaration(element);
      if(namespaceDeclaration == null) {
        return;
      }

      ICSharpNamespaceDeclaration csharpNamespaceDeclaration = namespaceDeclaration as ICSharpNamespaceDeclaration;
      if(csharpNamespaceDeclaration == null) {
        return;
      }

      string namespaceName = GetNamespace(projectFile);
      if(namespaceName == null) {
        return;
      }

      csharpNamespaceDeclaration.SetQualifiedName(namespaceName);
    }

    /// <summary>
    /// Gets the element as the caret position.
    /// </summary>
    /// <returns>The element.</returns>
    IElement GetElementAtCaret() {
      IProjectFile projectFile = DocumentManager.GetInstance(_solution).GetProjectFile(_textControl.Document);
      if(projectFile == null){
        return null;
      }

      ICSharpFile file = PsiManager.GetInstance(_solution).GetPsiFile(projectFile) as ICSharpFile;
      if(file == null){
        return null;
      }

      return file.FindTokenAt(_textControl.CaretModel.Offset);
    }

    #endregion
  }
}