using System;
using System.Collections.Generic;
using System.IO;
using System.Text.RegularExpressions;
using JetBrains.ReSharper.Daemon;
using JetBrains.ReSharper.Editor;
using JetBrains.ReSharper.ProjectModel;
using JetBrains.ReSharper.Psi.CSharp.Tree;
using JetBrains.ReSharper.Psi.Tree;
using JetBrains.ReSharper.TextControl;
using JetBrains.ReSharper.Util;
using ReSharper.Framework;

namespace ReSharper.SynchronizeNamespaceWithFilePath {
  /// <summary>
  /// Implements a Generate Assertions powertoy plugin.
  /// </summary>
  [ContextAction(Name = "SynchronizeNamespaceWithFilePath")]
  public class ContextAction : ContextActionBase {
    #region Constructor

    /// <summary>
    /// Initializes a new instance of the <see cref="ContextAction"/> class.
    /// </summary>
    /// <param name="solution">The solution.</param>
    /// <param name="textControl">The text control.</param>
    public ContextAction(ISolution solution, ITextControl textControl) : base(solution, textControl) {
    }

    #endregion

    #region Public methods

    /// <summary>
    /// Gets the file path.
    /// </summary>
    /// <param name="project">The project.</param>
    /// <param name="projectFile">The project file.</param>
    /// <returns>The file path.</returns>
    public static FileSystemPath GetFilePath(IProject project, IProjectFile projectFile) {
      FileSystemPath filePath = projectFile.Location;

      filePath = filePath.ConvertToRelativePath(project.Location);

      return filePath.Directory;
    }

    /// <summary>
    /// Gets the namespace declaration.
    /// </summary>
    /// <param name="element">The element.</param>
    /// <returns>The namespace declaration.</returns>
    public static INamespaceDeclaration GetNamespaceDeclaration(IElement element) {
      ITreeNode node = element as ITreeNode;
      if(node == null) {
        return null;
      }

      while(node != null) {
        IOwnerQualificationNode qualificationNode = node as IOwnerQualificationNode;

        if(qualificationNode != null) {
          node = node.Parent;
          break;
        }

        ICSharpIdentifierNode identifierNode = node as ICSharpIdentifierNode;
        if(identifierNode != null) {
          node = node.Parent;

          while(node != null && node is IReferenceName) {
            node = node.Parent;
          }

          if(node is IOwnerQualificationNode) {
            node = node.Parent;
          }

          break;
        }

        node = node.Parent;
      }

      if(node == null) {
        return null;
      }

      INamespaceDeclaration namespaceDeclaration = node as INamespaceDeclaration;

      return namespaceDeclaration;
    }

    /// <summary>
    /// Determines whether the specified solution is available.
    /// </summary>
    /// <param name="solution">The solution.</param>
    /// <param name="projectFile">The project file.</param>
    /// <param name="element">The element.</param>
    /// <returns>
    /// 	<c>true</c> if the specified solution is available; otherwise, <c>false</c>.
    /// </returns>
    public static bool IsAvailable(ISolution solution, IProjectFile projectFile, IElement element) {
      INamespaceDeclaration namespaceDeclaration = element as INamespaceDeclaration;
      if(namespaceDeclaration == null) {
        return false;
      }

      ICSharpNamespaceDeclarationNode namespaceDeclarationNode = element as ICSharpNamespaceDeclarationNode;
      if(namespaceDeclarationNode == null) {
        return false;
      }

      IProject project = projectFile.GetProject();
      if(project == null) {
        return false;
      }

      FileSystemPath filePath = GetFilePath(project, projectFile);

      List<string> fileParts = new List<string>(filePath.GetPathComponents());
      if(fileParts.Count <= 0) {
        return false;
      }

      fileParts[fileParts.Count - 1] = Path.GetFileNameWithoutExtension(fileParts[fileParts.Count - 1]);

      for(int n = 0; n < fileParts.Count; n++){
        fileParts[n] = Regex.Replace(fileParts[n], "[^0-9a-zA-Z]", string.Empty);
      }

      for(int n = fileParts.Count - 1; n >= 0; n--){
        if(string.IsNullOrEmpty(fileParts[n])){
          fileParts.RemoveAt(n);
        }
      }

      List<string> namespaceParts = GetNamespace(project, namespaceDeclaration);
      if(namespaceParts == null){
        return false;
      }

      if(fileParts.Count != namespaceParts.Count){
        return true;
      }

      for(int n = 0; n < fileParts.Count; n++){
        if(string.Compare(fileParts[n], namespaceParts[n], true) != 0){
          return true;
        }
      }

      return false;
    }

    /// <summary>
    /// Gets the namespace.
    /// </summary>
    /// <param name="namespaceDeclaration">The namespace declaration.</param>
    /// <param name="project">The project.</param>
    /// <returns>The namespace.</returns>
    public static List<string> GetNamespace(IProject project, INamespaceDeclaration namespaceDeclaration) {
      string namespacePath = namespaceDeclaration.DeclaredName;
      if(string.IsNullOrEmpty(namespacePath)) {
        return null;
      }

      List<string> namespaceParts = new List<string>(namespacePath.Split('.'));

      FileSystemPath firstFolder = project.Location.Combine(namespaceParts[0]);

      if(!Directory.Exists(firstFolder.ToString())){
        namespaceParts.RemoveAt(0);
      }

      return namespaceParts;
    }

    #endregion

    #region Protected methods

    /// <summary>
    /// Check if this action is available at the constructed context.
    /// Actions could store precalculated info in cache to share it between different actions
    /// </summary>
    /// <param name="element">The element.</param>
    /// <returns>
    /// 	<c>true</c> if available; otherwise, <c>false</c>.
    /// </returns>
    protected override bool IsAvailable(IElement element) {
      INamespaceDeclaration namespaceDeclaration = GetNamespaceDeclaration(element);
      if(namespaceDeclaration == null) {
        return false;
      }

      DocumentManager documentManager = DocumentManager.GetInstance(Solution);
      if(documentManager == null){
        return false;
      }

      IProjectFile projectFile = documentManager.GetProjectFile(TextControl.Document);
      if(projectFile == null) {
        return false;
      }

      return IsAvailable(Solution, projectFile, namespaceDeclaration);
    }

    /// <summary>
    /// Executes this instance.
    /// </summary>
    /// <param name="element">The element.</param>
    protected override void Execute(IElement element) {
    }

    /// <summary>
    /// Gets the items.
    /// </summary>
    /// <returns>The items.</returns>
    protected override IBulbItem[] GetItems() {
      List<IBulbItem> list = new List<IBulbItem>();

      GetBulbItems(list);

      return list.ToArray();
    }

    /// <summary>
    /// Gets the bulb items.
    /// </summary>
    /// <param name="list">The list.</param>
    void GetBulbItems(List<IBulbItem> list) {
      list.Add(new MoveFile(Solution, TextControl));

      DocumentManager documentManager = DocumentManager.GetInstance(Solution);
      if(documentManager == null){
        return;
      }

      IProjectFile projectFile = documentManager.GetProjectFile(TextControl.Document);

      if(projectFile == null){
        return;
      }

      IProject project = projectFile.GetProject();

      if(project == null){
        return;
      }

      FileSystemPath filePath = GetFilePath(project, projectFile);

      string path = filePath.ToString();

      if(string.IsNullOrEmpty(path) || path == "/"){
        return;
      }

      list.Add(new ChangeNamespace(Solution, TextControl));
    }

    /// <summary>
    /// Gets the text.
    /// </summary>
    /// <returns>The text.</returns>
    protected override string GetText() {
      return string.Empty;
    }

    #endregion
  }
}
  

