using System;
using System.IO;
using EnvDTE;
using JetBrains.ReSharper.Daemon;
using JetBrains.ReSharper.Editor;
using JetBrains.ReSharper.ProjectModel;
using JetBrains.ReSharper.Psi;
using JetBrains.ReSharper.Psi.CSharp.Tree;
using JetBrains.ReSharper.Psi.Tree;
using JetBrains.ReSharper.Shell;
using JetBrains.ReSharper.TextControl;
using JetBrains.ReSharper.Util;
using JetBrains.VSAddin.ProjectModel;

namespace ReSharper.SynchronizeNamespaceWithFilePath {
  /// <summary>
  /// Implements a Generate Assertions powertoy plugin.
  /// </summary>
  public class MoveFile : IBulbItem {
    #region Fields

    readonly ISolution _solution;
    readonly ITextControl _textControl;

    #endregion

    #region Constructor

    /// <summary>
    /// Initializes a new instance of the <see cref="ContextAction"/> class.
    /// </summary>
    /// <param name="solution">The solution.</param>
    /// <param name="textControl">The text control.</param>
    public MoveFile(ISolution solution, ITextControl textControl) {
      _solution = solution;
      _textControl = textControl;
    }

    #endregion

    #region Public properties

    /// <summary>
    /// Gets the text.
    /// </summary>
    /// <value>The text.</value>
    public string Text {
      get {
        IProjectFile projectFile = DocumentManager.GetInstance(_solution).GetProjectFile(_textControl.Document);
        if(projectFile == null) {
          return "Move file";
        }

        return "Move file to folder " + GetTargetFile(projectFile, true);
      }
    }

    #endregion

    #region Public methods

    /// <summary>
    /// Executes the action.
    /// </summary>
    /// <param name="solution">The solution.</param>
    /// <param name="textControl">The text control.</param>
    public void Execute(ISolution solution, ITextControl textControl) {
      if(_solution != solution || _textControl != textControl){
        throw new InvalidOperationException();
      }

      Shell.Instance.AssertReadAccessAllowed();

      IProjectFile projectFile = DocumentManager.GetInstance(_solution).GetProjectFile(_textControl.Document);
      if(projectFile == null) {
        return;
      }

      ProjectModelSynchronizer synchronizer = ProjectModelSynchronizer.GetInstance(projectFile.GetSolution());
      if(synchronizer == null){
        return;
      }

      VSProjectInfo projectInfo = synchronizer.GetProjectInfoByProject(projectFile.GetProject());
      if(projectInfo == null) {
        return;
      }

      Project vsProject = projectInfo.GetExtProject();
      if(vsProject == null) {
        return;
      }

      DTE dte = vsProject.DTE;
      if(dte == null){
        return;
      }

      Solution dteSolution = dte.Solution;
      if(dteSolution == null){
        return;
      }

      string sourceFile = projectFile.Location.ToString();

      ProjectItem item = dteSolution.FindProjectItem(sourceFile);
      if(item == null){
        return;
      }

      string targetFile = GetTargetFile(projectFile, false);
      if(targetFile == null){
        return;
      }

      Directory.CreateDirectory(Path.GetDirectoryName(targetFile));

      if(item.SaveAs(targetFile)){
        File.Delete(sourceFile);
      }
    }

    /// <summary>
    /// Gets the target file.
    /// </summary>
    /// <param name="projectFile">The project file.</param>
    /// <param name="relative">if set to <c>true</c> [relative].</param>
    /// <returns>The target file.</returns>
    public string GetTargetFile(IProjectFile projectFile, bool relative) {
      IProject project = projectFile.GetProject();
      if(project == null) {
        return null;
      }

      IElement element = GetElementAtCaret();
      if(element == null) {
        return null;
      }

      INamespaceDeclaration namespaceDeclaration = ContextAction.GetNamespaceDeclaration(element);
      if(namespaceDeclaration == null) {
        return null;
      }

      ICSharpNamespaceDeclarationNode namespaceDeclarationNode = namespaceDeclaration as ICSharpNamespaceDeclarationNode;
      if(namespaceDeclarationNode == null){
        return null;
      }

      FileSystemPath folder = new FileSystemPath(namespaceDeclarationNode.QualifiedName.Replace('.', '\\'));

      string[] parts = folder.GetPathComponents();

      if(parts.Length > 0) {
        FileSystemPath firstFolder = project.Location.Combine(parts[0]);

        if(!Directory.Exists(firstFolder.ToString())) {
          folder = folder.RemoveFirstComponent();
        }
      }

      folder = project.Location.Combine(folder);

      folder = folder.Combine(Path.GetFileName(projectFile.Location.ToString()));

      if (!relative){
        return folder.ToString();
      }

      folder = folder.ConvertToRelativePath(project.Location);

      return "\\" + Path.GetDirectoryName(folder.ToString());
    }

    #endregion

    #region Private methods

    /// <summary>
    /// Gets the element as the caret position.
    /// </summary>
    /// <returns>The element.</returns>
    IElement GetElementAtCaret() {
      IProjectFile projectFile = DocumentManager.GetInstance(_solution).GetProjectFile(_textControl.Document);
      if(projectFile == null){
        return null;
      }

      ICSharpFile file = PsiManager.GetInstance(_solution).GetPsiFile(projectFile) as ICSharpFile;
      if(file == null){
        return null;
      }

      return file.FindTokenAt(_textControl.CaretModel.Offset);
    }

    #endregion
  }
}
